/* ============================================================================
 * DCOPFJ : An open-source Java solver for DC optimal power flow (OPF) problems
 * ============================================================================
 *
 * (C) Copyright 2007, by Junjie Sun and Leigh Tesfatsion
 *
 * Version 2.0
 *
 * Project site:  http://www.econ.iastate.edu/tesfatsi/DCOPFJHome.htm
 *
 * LICENSING TERMS
 * The DCOPFJ package is licensed by the copyright holders (Junjie Sun and 
 * Leigh Tesfatsion) as free open-source software under the terms of the GNU
 * General Public License (GPL). Anyone who is interested is allowed to view, 
 * modify, and/or improve upon the code used to produce this package, but any 
 * software generated using all or part of this code must be released as free 
 * open-source software in turn. The GNU GPL can be viewed in its entirety as 
 * in the following site: http://www.gnu.org/licenses/gpl.html
 */

package dcopfj;

// DCOPFJ implementation Version 2.0
// DCOPFJ.java
// DC Optimal Power Flow (OPF) solver
//
// DCOPFJ formulates the needed matrices and vectors for QuadProgJ
//
// DC-OPF paper: Junjie Sun and Leigh Tesfatsion, (2006) "DC OPF Formulation
//     and Solution Using QuadProgJ", ISU Econ Working Paper Series #06014

import cern.colt.matrix.*;
import cern.colt.matrix.impl.*;
import quadprogj.*;

/**
 * What's new: this version of DCOPFJ can handle both case of fixed demands
 * (as in the old version) and price-sensitive demands (as in this version). 
 *  
 *      LSE demand  = fixed demand
 *   OR
 *      LSE demand = fixed demand + price-sensitive demand bid function
 * 
 * where fixed demand is simply the hourly load profiles, and the price-sensitive
 * demand bid function for each LSE j is expressed as follows:
 *
 *      D_j(sLoad_j) = c_j - 2 * d_j * sLoad_j
 *
 * where sLoad is the price-sensitive load demand quantity
 *
 */

/**
 * Conform the DC-OPF problem to a strictly convex QP problem.  See DC-OPF paper
 *
 * To form the input matrices/vector for QuadProgJ, we need the following internal
 * variable names for two cases: fixed demand and price-sensitive demand:
 *=======================================================================
 * CASE 1: FIXED DEMAND
 *-------------------------------
 * G:     U, Wrr
 * a:     A_1...A_I
 * Ceq:   II, rBusAdm
 * beq:   FDemand
 * Ciq:   Oni, Z, rAdj, Iii, Oik
 * biq:   pU, capL, capU
 * ------------------------------
 * B:       supplyOffer[][1]; cost coefficient b; as appeared in matrix U
 * A:       supplyOffer[][0]; cost coefficient a
 * Wrr:     grid.getReducedVADWeight() 
 * II:      someFunction(atNodeByGen)
 * rBusAdm: grid.getReducedBusAdmittance()
 * FDemand: someFunction(atNodeByLSE, loadProfile[]) ~ Section 4.3 in DC-OPF paper
 * Oni:     NxI zero matrix
 * Z:       grid.getDiagonalAdmittance() 
 * rAdj:    grid.getReducedAdjacency() 
 * Iii:     IxI identity matrix
 * Oik:     Ix(K-1) zero matrix
 * pU:      grid.getLineCap()
 * capL:    supplyOffer[][2]; capL; Generator's lower operating capacity limit
 * capU:    supplyOffer[][3]; capU; Generator's upper operating capacity limit
 *=======================================================================
 * CASE 2: PRICE-SENSITIVE DEMAND
 *-------------------------------
 * G:     U, Wrr
 * a:     A_1...A_I, C_1...C_J
 * Ceq:   II, JJ, rBusAdm
 * beq:   FDemand
 * Ciq:   Oni, Onj, Z, rAdj; Iii, Oij, Oik; Oji, Ijj, Ojk
 * biq:   pU, capL, capU, sLoadL, sLoadU
 * ------------------------------
 * U:       B_1...B_I, D_1...D_J
 * B:       supplyOffer[][1]; cost coefficient b; as appeared in matrix U
 * D:       psDemandBid[][1]; demand coefficient d; as appeared in matrix U
 * A:       supplyOffer[][0]; cost coefficient a
 * C:       psDemandBid[][0]; demand coefficient c
 * Wrr:     grid.getReducedVADWeight() 
 * II:      someFunction(atNodeByGen)
 * JJ:      someFunction(atNodeByLSE)
 * rBusAdm: grid.getReducedBusAdmittance()
 * FDemand: someFunction(atNodeByLSE, loadProfile[]) ~ Section 4.3 in DC-OPF paper
 * Oni:     NxI zero matrix
 * Onj:     NxJ zero matrix
 * Z:       grid.getDiagonalAdmittance() 
 * rAdj:    grid.getReducedAdjacency() 
 * Iii:     IxI identity matrix
 * Oij:     IxJ zero matrix
 * Oik:     Ix(K-1) zero matrix
 * Oji:     JxI zero matrix
 * Ijj:     JxJ identity matrix
 * Ojk:     JxK zero matrix
 * pU:      grid.getLineCap()
 * capL:    supplyOffer[][2]; capL; Generator's lower operating capacity limit
 * capU:    supplyOffer[][3]; capU; Generator's upper operating capacity limit
 * sLoadL:  psDemandBid[][2]; sLoadL; LSE's price-sensitive lower load limit
 * sLoadU:  psDemandBid[][3]; sLoadU; LSE's price-sensitive upper load limit
 */
public class DCOPFJ {

  // Index for supplyOffer parameters, i.e., in the form of {a,b,capL,capU}
  private static final int A_INDEX    = 0;
  private static final int B_INDEX    = 1;
  private static final int CAP_LOWER  = 2;
  private static final int CAP_UPPER  = 3;

  // Index for psDemandBid parameters, i.e., in the form of {c,d,sLoadL,sLoadU}
  private static final int C_INDEX   = 0;
  private static final int D_INDEX   = 1;
  private static final int SLOAD_LOWER    = 2;
  private static final int SLOAD_UPPER    = 3;

  private DoubleFactory1D fac1d = DoubleFactory1D.dense;
  private DoubleFactory2D fac2d = DoubleFactory2D.dense;
    // for using Colt's methods e.g. diagonal(), identity(), etc.
  private cern.jet.math.Functions F = cern.jet.math.Functions.functions;
    // F: Naming shortcut to save some keystrokes for calling Colt's functions

  // Input data from BSCUC
  private DoubleMatrix2D supplyOffer;
  private DoubleMatrix2D psDemandBid; //psDemandBid = price-sensitive demand bid
  private double[] loadProfile; // loadProfile = fixed demand
  private int[] atNodeByGen;
  private int[] atNodeByLSE;
  private TransGrid grid;

  // Intermediate input to form (G,a,Ceq,beq,Ciq,biq)
  private int K;               // numNodes
  private int N;               // numBranches
  private int I;               // numGenAgents
  private int J;               // numLSEAgents
  private DoubleMatrix1D B;    // Ix1
  private DoubleMatrix2D U;    // Ix1
  private DoubleMatrix2D Wrr;  // (K-1)x(K-1)
  private DoubleMatrix1D A;    // Ix1
  private DoubleMatrix2D II;  // KxI
  private DoubleMatrix2D rBusAdm;   // (K-1)xK
  private DoubleMatrix1D FDemand;  // Kx1
  private DoubleMatrix2D Oni;   // NxI
  private DoubleMatrix2D Z;    // NxN
  private DoubleMatrix2D rAdj; // Nx(K-1)
  private DoubleMatrix2D Iii;   // IxI
  private DoubleMatrix2D Oik;   // Ix(K-1)
  private DoubleMatrix1D pU;   // Nx1
  private DoubleMatrix1D capL;  // Ix1
  private DoubleMatrix1D capU;  // Ix1
  //Below are the additional variables for the price-sensitive demand case
  private DoubleMatrix1D D; // Jx1
  private DoubleMatrix1D C; // Jx1
  private DoubleMatrix2D JJ; // KxJ 
  private DoubleMatrix2D Onj; // NxJ
  private DoubleMatrix2D Oij; // IxJ
  private DoubleMatrix2D Oji; // JxI
  private DoubleMatrix2D Ijj; // JxJ
  private DoubleMatrix2D Ojk; // JxK
  private DoubleMatrix1D sLoadL; // Jx1
  private DoubleMatrix1D sLoadU; // Jx1
  
  // Input for QuadProgJ
  private DoubleMatrix2D G;
  private DoubleMatrix1D a;
  private DoubleMatrix2D Ceq;
  private DoubleMatrix1D beq;
  private DoubleMatrix2D Ciq;
  private DoubleMatrix1D biq;

  private QuadProgJ qpj;

  // Solution from QuadProgJ
  private double[] commitment; // power production quantity
  private double[] voltAngle;  // voltage angle in radians
  private double[] voltAngleDegree;  // voltage angle in degrees
  private double[] lmp;  // lmp (locational marginal prices) == eqMultiplier
  private double[] ineqMultiplier;
  private double minTVC; // minTVC = SUM (Ai*PGi + Bi*PGi^2)
  private double[] branchFlow; // branchFlow_km = (1/x_km)(delta_k - delta_m)
  private double[][] bi; // branch index
  private double sumSquaredAngleDifference;  // SUM(delta_k - delta_m)^2
  private int demandType; //demandType == 0 in fixed dmeand; == 1 in price-sensitive demand
  private double[] sLoad; //price-sensitive load demand quantity

  // Constructor for fixed demand
  public DCOPFJ(double[][] so, double[] lp, int[] ng, int[] nl, TransGrid tg){
    supplyOffer = new DenseDoubleMatrix2D(so);   // Ix4
    loadProfile = lp;  // Jx1
    atNodeByGen = ng;  // Ix1
    atNodeByLSE = nl;  // Jx1
    grid        = tg;

    K = grid.getNumNodes();
    N = grid.getNumBranches();
    I = atNodeByGen.length;
    J = atNodeByLSE.length;
    
    demandType = 0; // demandType = 0 if it is fixed demand; 
                    // demandType = 1 if it is price-sensitive demand.
    solveDCOPF(demandType);

  }

  // Constructor for price-sensitive demand
  public DCOPFJ(double[][] so, double[][] db, double[] lp, int[] ng, int[] nl, 
      TransGrid tg){
    supplyOffer = new DenseDoubleMatrix2D(so);// Ix4; supply offer
    psDemandBid = new DenseDoubleMatrix2D(db);// Jx4; price-sensitive demand bid
    loadProfile = lp;  // Jx1; fixed demand
    atNodeByGen = ng;  // Ix1; Gen location on the grid
    atNodeByLSE = nl;  // Jx1; LSE location on the grid
    grid        = tg;  // Grid data

    K = grid.getNumNodes();
    N = grid.getNumBranches();
    I = atNodeByGen.length;
    J = atNodeByLSE.length;

    demandType = 1; // demandType = 0 if it is fixed demand; 
                    // demandType = 1 if it is price-sensitive demand.
    solveDCOPF(demandType);

  }
  
  
  private void solveDCOPF(int demandType){
    int dt = demandType;  // dt = 0 if fixed demand; dt = 1 if price-sensitive demand
    formG(dt);
    forma(dt);
    formCeq(dt);
    formbeq();
    formCiq(dt);
    formbiq(dt);
    /*
    System.out.println("G: " + G);
    System.out.println("a: " + a);
    System.out.println("Ceq': " + Ceq.viewDice());
    System.out.println("beq: " + beq);
    System.out.println("Ciq': " + Ciq.viewDice());
    System.out.println("biq: " + biq); */

    qpj = new QuadProgJ(G,a,Ceq,beq,Ciq,biq);

    commitment = new double[I];   // in MWs
    voltAngle = new double[K-1];  // in radians
    voltAngleDegree = new double[K-1]; // in degress
    lmp        = new double[K];
    ineqMultiplier = new double[2*N+2*I];
    minTVC = 0;  // in $/h
    branchFlow = new double[N]; //in MWs
    bi = new double[N][2]; // columns are FROM and TO; rows are branches
    bi = grid.getBranchIndex();
    double[] fullVoltAngle = new double[K]; // including delta_1 = 0
    sumSquaredAngleDifference = 0;
    

    // NOTE FOR THE SOLUTION STRUCTURE x* = qpj.getMinX()
    // x* = (p_{G1}...p_{GI}, delta_2...delta_K) for fixed demand
    // x* = (p_{G1}...p_{GI}, p_{L1}^S...p_{LJ}^S, delta_2...delta_K) 
    //      for price-sensitive demand
    
    // DC-OPF solution for (p_{G1},...,p_{GI}) in SI
    for(int i=0; i<I; i++){
      commitment[i] = qpj.getMinX()[i]*INIT.getBaseS();
    }
    if (demandType == 0){
      // DC-OPF solution for (delta_2,...,delta_K)
      for(int k=I; k<I+K-1; k++){
        voltAngle[k-I] = qpj.getMinX()[k];  // voltAngle in radians
      }
    }
    else if (demandType == 1){
      sLoad = new double[J];
      // DC-OPF solution for (p_{L1}^S,...,p_{LJ}^S) in SI
      for(int j=I; j<I+J; j++){
        sLoad[j-I] = qpj.getMinX()[j]*INIT.getBaseS();
      }
      // DC-OPF solution for (delta_2,...,delta_K)
      for(int k=I+J; k<I+J+K-1; k++){
        voltAngle[k-I-J] = qpj.getMinX()[k];  // voltAngle in radians
      }   
    }
    else{
      System.out.println("Wrong demand type!");
    }
    
    // Convert voltage angle from randian to degree
    for(int k=1; k<K-1; k++){
      voltAngleDegree[k] = (voltAngle[k]*180)/Math.PI; // volt angle in degrees
    }

    // lmp: locational marginal prices in SI
    for(int k=0; k<K; k++){
      lmp[k] = qpj.getEqMultipliers()[k]/INIT.getBaseS();
    }
    
    for(int j=0; j<2*N+2*I; j++){
      ineqMultiplier[j] = qpj.getIneqMultipiers()[j]/INIT.getBaseS();
    }
    for(int i=0; i<I; i++){
      minTVC = minTVC + (A.get(i)/INIT.getBaseS())*commitment[i]
      +(B.get(i)/(INIT.getBaseS()*INIT.getBaseS()))*commitment[i]*commitment[i];
    }

    for(int k=1; k<K; k++){
      fullVoltAngle[k] = voltAngle[k-1];
    }
    //System.out.println("fullVoltAngle");
    //Support.print(fullVoltAngle,4);
    for(int n=0; n<N; n++){
      //System.out.println("1/x_km: " +(1/grid.getReactance()[n]));
      //System.out.println("k,m: " + bi[n][0] + "," + bi[n][1]);
      //System.out.println("delta_k: " + fullVoltAngle[(int)bi[n][0]-1]);
      //System.out.println("delta_m: " + fullVoltAngle[(int)bi[n][1]-1]);
      branchFlow[n] = (1/grid.getReactance()[n])*(fullVoltAngle[(int)bi[n][0]-1]
                      - fullVoltAngle[(int)bi[n][1]-1])*INIT.getBaseS();

      sumSquaredAngleDifference = sumSquaredAngleDifference
          + Math.pow((fullVoltAngle[(int)bi[n][0]-1]
                      - fullVoltAngle[(int)bi[n][1]-1]),2);
    }

  }

  // G = blockDiag(U,Wr), where U = diag(2B) or U = diag(2B, 2D)
  private void formG(int dt){
    int demandType = dt;
    B = new DenseDoubleMatrix1D(supplyOffer.viewColumn(B_INDEX).toArray());
    if (demandType == 0){
      U = new DenseDoubleMatrix2D(fac2d.diagonal(B.assign(F.mult(2))).toArray());
      G = new DenseDoubleMatrix2D(I+K-1,I+K-1);
    }
    else if (demandType == 1){
      D = new DenseDoubleMatrix1D(psDemandBid.viewColumn(D_INDEX).toArray());
      U = new DenseDoubleMatrix2D
         (fac2d.diagonal(fac1d.append(B,D).assign(F.mult(2))).toArray());
      G = new DenseDoubleMatrix2D(I+J+K-1,I+J+K-1);
    }
    else{
      System.out.println("Wrong demand type; demand type = 0 if fixed demand" 
          + "demand type = 1 if price-sensitive demand.");
    }
    Wrr = new DenseDoubleMatrix2D(grid.getReducedVADWeight());
    G.assign(fac2d.composeDiagonal(U,Wrr));
  }

  // a = (A, 0...0) or a = (A,-C,0...0)
  private void forma(int dt){
    int demandType = dt;
    A = new DenseDoubleMatrix1D(supplyOffer.viewColumn(A_INDEX).toArray());
    if (demandType == 0){
      a = new DenseDoubleMatrix1D(I+K-1);
      a.viewPart(0,I).assign(A); // viewPart(fromIndex, width)
    }
    else if (demandType == 1){
      C = new DenseDoubleMatrix1D(psDemandBid.viewColumn(C_INDEX).toArray());
      a = new DenseDoubleMatrix1D(I+J+K-1);
      a.viewPart(0,I).assign(A);
      a.viewPart(I,J).assign(C.assign(F.neg));
    }
    else{
      System.out.println("Wrong demand type; demand type = 0 if fixed demand" 
          + "demand type = 1 if price-sensitive demand.");
    }        
  }

  // CeqTranspose = (II, -Br'); Ceq = CeqTranspose'; where Br' is rBusAdm here
  private void formCeq(int dt){
    int demandType = dt;
    II = new DenseDoubleMatrix2D(K,I);
    for(int k=0; k<K; k++){
      for(int i=0; i<I; i++){
        if(atNodeByGen[i]==k+1){
          II.set(k,i,1);  // change set to setQuick() later
        }
      }
    }
    rBusAdm = new DenseDoubleMatrix2D(grid.getReducedBusAdmittance());
    
    if (demandType == 0){
      DoubleMatrix2D[][] parts = {{ II, rBusAdm.viewDice().assign(F.neg)}};
      DoubleMatrix2D CeqTranspose = new DenseDoubleMatrix2D(K,I+K-1);
      CeqTranspose.assign(fac2d.compose(parts));
      Ceq = new DenseDoubleMatrix2D(I+K-1,K);
      Ceq.assign(CeqTranspose.viewDice());
    }
    else if (demandType == 1){
      JJ = new DenseDoubleMatrix2D(K,J);
      for(int k=0; k<K; k++){
        for(int j=0; j<J; j++){
          if(atNodeByLSE[j]==k+1){
            JJ.set(k,j,1);  // change set to setQuick() later
          }
        }
      }
     DoubleMatrix2D[][] parts = {{ II, JJ.assign(F.neg),
         rBusAdm.viewDice().assign(F.neg)}};
     DoubleMatrix2D CeqTranspose = new DenseDoubleMatrix2D(K,I+J+K-1);
     CeqTranspose.assign(fac2d.compose(parts));
     Ceq = new DenseDoubleMatrix2D(I+J+K-1,K);
     Ceq.assign(CeqTranspose.viewDice());          
    }
    else{
      System.out.println("Wrong demand type; demand type = 0 if fixed demand" 
          + "demand type = 1 if price-sensitive demand.");
    }          

  }
  // FDemand = someFunction(atNodeByLSE, loadProfile); beq = FDemand
  private void formbeq(){
    FDemand = new DenseDoubleMatrix1D(K);
    for(int k=0; k<K; k++){
      double lp=0;
      for(int j=0; j<J; j++){
        if(atNodeByLSE[j]==k+1){
          lp = lp + loadProfile[j];
        }
      }
      FDemand.set(k,lp);
    }
    beq = new DenseDoubleMatrix1D(K);
    beq.assign(FDemand);
  }
  
  // Ciq matrix formulation
  //
  // (1) FIXED DEMAND CASE:
  // CiqTranspose = {{Oni, Z*rAdj},{Oni, -Z*rAdj},{Iii, Oik},{-Iii, Oik}};
  // Ciq = CiqTranspose'
  //
  // (2) PRICE-SENSITIVE DEMAND CASE:
  // CiqTranspose = {{MatrixT}, {MatrixG}, {MatrixL}};
  // MatrixT = {{Oni, Onj, Z*rAdj}, {Oni, Onj, -Z*rAdj}};
  // MatrixG = {{Iii, Oij, Oik},{-Iii, Oij, Oik}};
  // MatrixL = {{Oji, Ijj, Ojk},{Oji, -Ijj, Ojk}};
  private void formCiq(int dt){
    int demandType = dt;
    Oni   = new DenseDoubleMatrix2D(N,I);
    Z    = new DenseDoubleMatrix2D(grid.getDiagonalAdmittance());
    rAdj = new DenseDoubleMatrix2D(grid.getReducedAdjacency());
    Iii = new DenseDoubleMatrix2D(I,I).assign(fac2d.identity(I));
    Oik = new DenseDoubleMatrix2D(I,K-1);
    
    if (demandType == 0){
      DoubleMatrix2D[][] parts = {
      { Oni,                      Z.zMult(rAdj,null)                     },
      { Oni,                      Z.copy().assign(F.neg).zMult(rAdj,null)},
      { Iii,                      Oik                                    },
      { Iii.copy().assign(F.neg), Oik                                    }};
      DoubleMatrix2D CiqTranspose = new DenseDoubleMatrix2D(2*N+2*I,I+K-1);
      CiqTranspose.assign(fac2d.compose(parts));
      Ciq = new DenseDoubleMatrix2D(I+K-1,2*N+2*I);
      Ciq.assign(CiqTranspose.viewDice()); 
    }
    else if (demandType == 1){
      Onj = new DenseDoubleMatrix2D(N,J);
      Oij = new DenseDoubleMatrix2D(I,J);
      Oji = new DenseDoubleMatrix2D(J,I);
      Ijj = new DenseDoubleMatrix2D(J,J).assign(fac2d.identity(J));
      Ojk = new DenseDoubleMatrix2D(J,K-1);
      DoubleMatrix2D[][] parts = {
        { Oni,                      Onj, Z.zMult(rAdj,null)},
        { Oni,                      Onj, Z.copy().assign(F.neg).zMult(rAdj,null)},
        { Iii,                      Oij, Oik},
        { Iii.copy().assign(F.neg), Oij, Oik},
        { Oji,                      Ijj, Ojk},
        { Oji,                      Ijj.copy().assign(F.neg), Ojk}};
      DoubleMatrix2D CiqTranspose = new DenseDoubleMatrix2D(2*N+2*I+2*J,I+J+K-1);
      CiqTranspose.assign(fac2d.compose(parts));
      Ciq = new DenseDoubleMatrix2D(I+J+K-1,2*N+2*I+2*J);
      Ciq.assign(CiqTranspose.viewDice());       
    }
    else{
      System.out.println("Wrong demand type; demand type = 0 if fixed demand" 
          + "demand type = 1 if price-sensitive demand.");
    }            
  }
  
  // biq = (-pU, -pU, capL, -capU) or biq = (-pU, -pU, capL, -capU, sLoadL, -sLoadU)
  private void formbiq(int dt){
    int demandType = dt;
    pU = new DenseDoubleMatrix1D(grid.getLineCap());
    capL = new DenseDoubleMatrix1D(supplyOffer.viewColumn(CAP_LOWER).toArray());
    capU = new DenseDoubleMatrix1D(supplyOffer.viewColumn(CAP_UPPER).toArray());

    if (demandType == 0){
      DoubleMatrix1D[] parts = {pU.copy().assign(F.neg), 
                pU.copy().assign(F.neg), capL, capU.copy().assign(F.neg)};
      biq = new DenseDoubleMatrix1D(2*N+2*I);
      biq.assign(fac1d.make(parts));
    }
    else if (demandType == 1){
      sLoadL = new DenseDoubleMatrix1D(psDemandBid.viewColumn(SLOAD_LOWER).toArray());
      sLoadU = new DenseDoubleMatrix1D(psDemandBid.viewColumn(SLOAD_UPPER).toArray());
      DoubleMatrix1D[] parts = 
        {pU.copy().assign(F.neg), pU.copy().assign(F.neg), 
         capL,                    capU.copy().assign(F.neg),
         sLoadL,                  sLoadU.copy().assign(F.neg)};
      biq = new DenseDoubleMatrix1D(2*N+2*I+2*J);
      biq.assign(fac1d.make(parts));
      
    }
    else{
      System.out.println("Wrong demand type; demand type = 0 if fixed demand" 
          + "demand type = 1 if price-sensitive demand.");
    }                
  }

  public double[] getCommitment(){
    return commitment;
  }
  public double[] getSLoad(){
    if (demandType == 1){
      return sLoad;
    }
    else{
      System.out.println("Wrong request! sLoad only exisits in price-sensitive demand case");
      return null;
    }
  }
  public double[] getVoltAngle(){
    return voltAngle;
  }
  public double[] getVoltAngleDegree(){
  return voltAngleDegree;
 }
  public double[] getLMP(){
    return lmp;
  }
  public double[] getIneqMultiplier(){
    return ineqMultiplier;
  }
  public double getMinTVC(){
    return minTVC;
  }
  public double[] getBranchFlow(){
    return branchFlow;
  }
  public double getSumSquaredAngleDifference(){
    return sumSquaredAngleDifference;
  }
  public int getNumBindingConstraints(){
    return qpj.getNumBC();
  }
  public int[] getActiveSet(){
    return qpj.getActiveSet();
  }
  public boolean getIsSolutionFeasibleAndOptimal(){
    return qpj.getIsFeasibleAndOptimal();
  }




  // DCOPFJ's main method
  public static void main(String[] args) {

    System.out.println("<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>\n");
    System.out.println("     DCOPFJ: An Open-Source Java Solver for DC Optimal Power Flow (OPF) Problems    \n");
    System.out.println("             (C) Copyright 2007, by Junjie Sun and Leigh Tesfatsion              ");
    System.out.println("                            Version Release: " + 2.0);    
    System.out.println("        Package Site: http://www.econ.iastate.edu/tesfatsi/DCOPFJHome.htm");
    System.out.println("         Licensed as free open-source software by the copyright holders ");
    System.out.println("   (Sun and Tesfatsion) under the terms of the GNU General Public License (GPL) \n");
    System.out.println("<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>\n");    
    
    String[] caseName = {"3NodeCase","5NodeFixedDemand","5NodePSDemand",
          "30NodeFixedDemand","30NodePSDemand"};
    // caseID: 0 = 3NodeCase; 
    //         1 = 5NodeFixedDemand; 
    //         2 = 5NodePSDemand; 
    //         3 = 30NodeFixedDemand; 
    //         4 = 30NodePSDemand.
    int caseID = 2;

    INIT init = new INIT("DATA/" + caseName[caseID] +".txt");
    System.out.println("\nCase name: " + caseName[caseID] + "\n");
    int demandType = init.getDemandType();

    TransGrid grid = new TransGrid(init.getNodeData(),init.getBranchData(),40,40);
    double[][] genData = init.getGenData();
    double[][] lseDataFDemand = init.getLSEDataFixedDemand();

    int K = grid.getNumNodes();
    int N = grid.getNumBranches();
    int I = genData.length;
    int J = lseDataFDemand.length;
    int H = 24;


    DCOPFJ opf;
    double[][] dailyCommitment = new double[H][I]; // (24 hours by row)
    double[][] dailyVoltAngle = new double[H][K];
    double[][] dailyVoltAngleDegree = new double[H][K];
    double[][] dailyLMP = new double[H][K];
    double[][] dailyIneqMultiplier = new double[H][2*N+2*I];
    double[] dailyMinTVC = new double [H];
    double[][] dailyBranchFlow = new double[H][N];
    double[] dailySumSquaredAngleDifference = new double[H];
    double[][] dailySLoad = new double[H][J]; // in price-sensitive demand case only

    double[][] loadProfile = new double[J][H];
    for(int j=0; j<J; j++){
      for(int h=0; h<H; h++){
        loadProfile[j][h] = lseDataFDemand[j][h+2];
      }
    }

    int[] atNodeByGen = new int[I];
    for(int i=0; i<I; i++){
      atNodeByGen[i] = (int) genData[i][1];
    }

    int[] atNodeByLSE = new int[J];
    for(int j=0; j<J; j++){
      atNodeByLSE[j] = (int) lseDataFDemand[j][1];
    }
    
    //supplyOfferByGen = [a, b, capL, capU]
    double[][] supplyOfferByGen = new double[I][4];
    for(int i=0; i<I; i++){
      for(int p=0; p<4; p++){
        supplyOfferByGen[i][p] = genData[i][p+3];
      }
    }
    //psDemandByLSE = [c, d, sLoadL, sLoadU]
    double[][] psDemandBidByLSE = new double[J][4];
    if(demandType == 1){
      double[][] lseDataSDemand = init.getLSEDataPriceSensitiveDemand();
      for(int j=0; j<J; j++){
        for(int p=0; p<4; p++){
          psDemandBidByLSE[j][p] = lseDataSDemand[j][p+2];
        }
      }
    }

    for (int h=0; h<H; h++){

      double[] hourlyLoadProfileByLSE = new double[J];
      for(int j=0; j<J; j++){
        hourlyLoadProfileByLSE[j] = loadProfile[j][h];
      }
      
      if(demandType == 0){
        opf = new DCOPFJ(supplyOfferByGen,hourlyLoadProfileByLSE,
                atNodeByGen, atNodeByLSE, grid);
        dailyCommitment[h] = opf.getCommitment();
        dailyVoltAngle[h] = opf.getVoltAngle();//report voltAngle in radians
        dailyLMP[h]        = opf.getLMP();
        dailyIneqMultiplier[h] = opf.getIneqMultiplier();
        dailyMinTVC[h] = opf.getMinTVC();
        dailyBranchFlow[h] = opf.getBranchFlow();
        dailySumSquaredAngleDifference[h] = opf.getSumSquaredAngleDifference();
      }
      else if(demandType == 1){
        opf = new DCOPFJ(supplyOfferByGen,psDemandBidByLSE, hourlyLoadProfileByLSE,
                atNodeByGen, atNodeByLSE, grid);
        dailyCommitment[h] = opf.getCommitment();
        dailySLoad[h] = opf.getSLoad();
        dailyVoltAngle[h] = opf.getVoltAngle();//report voltAngle in radians
        dailyLMP[h]        = opf.getLMP();
        dailyIneqMultiplier[h] = opf.getIneqMultiplier();
        dailyMinTVC[h] = opf.getMinTVC();
        dailyBranchFlow[h] = opf.getBranchFlow();
        dailySumSquaredAngleDifference[h] = opf.getSumSquaredAngleDifference();        
      }
      else{
        System.out.println("Wrong demand type!");
      }
    }

    // Extracting capL and capU from the 3rd and 4th column of supplyOfferByGen 
    //  (used to show if the lower or upper constraints are binding in solution)
    double[] capL = new double[I];
    double[] capU = new double[I];
    for(int i=0; i<I; i++){
      capL[i] = supplyOfferByGen[i][2]*INIT.getBaseS();
    }
    for(int i=0; i<I; i++){
      capU[i] = supplyOfferByGen[i][3]*INIT.getBaseS();
    }
    
    System.out.println("  Daily 24-hour generation commitments (in MWs) for each Generator");
    Support.print("Gen",I);
    Support.printWithContraints(dailyCommitment,2,capL,capU);
    
    if(demandType == 1){
      System.out.println("  Daily 24-hour price-sensitive demand bids (in MWs) for each LSE");
      Support.print("LSE",J);
      Support.print(dailySLoad,2);
    }
    System.out.println("  Daily 24-hour voltage angles (in radians) for each non-reference node");
    Support.printNonRef("Delta",K);
    Support.print(dailyVoltAngle,4);
  
    System.out.println("  Daily 24-hour locational marginal prices (in $/MWh) for each node");
    Support.print("LMP", K);
    Support.print(dailyLMP,2);
    System.out.println("  Daily 24-hour minimum total variable cost levels (in $/h) for the system");
    Support.print(dailyMinTVC,2);
    System.out.println("  Daily 24-hour branch flows (in MWs) for each branch, in the order of bi");
    Support.print("BFlow", N);
    Support.print(dailyBranchFlow,2);
    System.out.println("  Daily 24-hour inequality contraint multipliers, in the order of biq");
    Support.printArray2dp(dailyIneqMultiplier);
    System.out.println("  Daily 24-hour sums of squared voltage angle differences (in radians)");
    Support.printArray(dailySumSquaredAngleDifference);

  }

}
