/* ============================================================================
 * DCOPFJ : An open-source Java solver for DC optimal power flow (OPF) problems
 * ============================================================================
 *
 * (C) Copyright 2007, by Junjie Sun and Leigh Tesfatsion
 *
 * Version 2.0
 *
 * Project site:  http://www.econ.iastate.edu/tesfatsi/DCOPFJHome.htm
 *
 * LICENSING TERMS
 * The DCOPFJ package is licensed by the copyright holders (Junjie Sun and 
 * Leigh Tesfatsion) as free open-source software under the terms of the GNU
 * General Public License (GPL). Anyone who is interested is allowed to view, 
 * modify, and/or improve upon the code used to produce this package, but any 
 * software generated using all or part of this code must be released as free 
 * open-source software in turn. The GNU GPL can be viewed in its entirety as 
 * in the following site: http://www.gnu.org/licenses/gpl.html
 */

package dcopfj;

// INIT.java
// Initialization of input data, parse raw data into double arrays, convert data
// from SI to PU

import java.io.*;
import java.util.StringTokenizer;
import java.util.ArrayList;
import cern.colt.matrix.*;
import cern.colt.matrix.impl.*;

/**
* Example of initial data file format in SI
*
NAME	5NodeFixedDemand
UNIT	SI
BASE_S	100
BASE_V	10

// SI (International System of Units) such as MW
// PU (Per Unit)
//
// Source: Lally, John (2002) Financial Transmission Rights Auction Example, 
//         ISO New England

DemandType
PriceSensitiveNone

NodeData
//NN	penaltyCoeff
5	0.05

BranchData

//From	To	lineCap	reactance
1	2	250	0.0281  
1	4	150	0.0304  
1	5	400	0.0064
2	3	350	0.0108
3	4	240	0.0297
4	5	240	0.0297

GenData
//ID	atNode	FCost	a	b	capL	capU	initMoney
  1	1	160	14	0.005	0	110	10000
  2	1	190	15	0.006	0	100	10000
  3	3	280	25	0.010	0	520	10000
  4	4	100	30	0.012	0	200	10000
  5	5	240	10	0.007	0	600	10000

LSEDataFixedDemand
//ID	atNode	LP-00	LP-01	LP-02	LP-03	LP-04	LP-05	LP-06	LP-07 
  1	2	350	322.93	305.04	296.02	287.16	291.59	296.02	314.07
  2	3	300	276.80	261.47	253.73	246.13	249.93	253.73	269.20
  3	4	250	230.66	217.89	211.44	205.11	208.28	211.44	224.33

//ID	atNode	LP-08	LP-09	LP-10	LP-11	LP-12	LP-13	LP-14	LP-15
  1	2	358.86	394.80	403.82	408.25	403.82	394.80	390.37	390.37
  2	3	307.60	338.40	346.13	349.93	346.13	338.40	334.60	334.60
  3	4	256.33	282.00	288.44	291.61	288.44	282.00	278.83	278.83

//ID	atNode	LP-16	LP-17	LP-18	LP-19	LP-20	LP-21	LP-22	LP-23 
  1	2	408.25	448.62	430.73	426.14	421.71	412.69	390.37	363.46
  2	3	349.93	384.53	369.20	365.26	361.47	353.73	334.60	311.53
  3	4  	291.61	320.44	307.67	304.39	301.22	294.78	278.83	259.61

EndOfFile

// DemandType is PriceSensitiveNone if it is fixed demand only
// DemandType is PriceSensitiveAll if it is all price-sensitive demand
// NN: Number of Nodes
// penaltyCoeff: soft penalty coefficient pi
// lineCap: Thermal limits
// FCost,a,b: Generator's cost attributes
// capL,capU: Generator's operating capacity limits
// LP-H: load profile (fixed demand) at hour H, where H=00,01,...,23

NOTE: any line starts with // will be ignored when data are read in

 */

public class INIT {

  // Tolerance value (used in correctRoundingError())
  private static final double TOL = 1.0E-9;

  // Default values
  public static final int BASE_S = 100;
  public static final int BASE_V = 10;

  // Number of parameters (number of columns in the data file)
  public static final int NUM_NODE_PARAMS = 2;
    // numNodes, softPenaltyCoeff
  public static final int NUM_BRANCH_PARAMS = 4;
    // FromNode, ToNode, reactance, lineCap (Thermal Limits)
  public static final int NUM_GEN_PARAMS = 8;
    // genID, atWhichNode, FCost, a, b, capL, capU, initMoney
  public static final int NUM_LSE_3SEC_PARAMS = 2+8;
    // lseID, atWhichNode, LP-00,...,LP-07
  public static final int NUM_LSE_PARAMS = 2+24;
    // lseID, atWhichNode, LP-00,...,LP-23
  public static final int NUM_LSE_PARAMS_SLOAD = 6;
    // lseID, atWhichNode, c, d, sLoadL, sLoadU

  private String[][] paramNames = {
   {"DemandType"},
   {"NN","penaltyCoeff"},
   {"From","To","lineCap","reactance"},
   {"ID","atNode","FCost","a","b","capL","capU","initMoney"},
   {"ID","atNode","c","d","sLoadL","sLoadU"},
   {"ID","atNode","LP-00","LP-01","LP-02","LP-03","LP-04","LP-05","LP-06","LP-07",
                  "LP-08","LP-09","LP-10","LP-11","LP-12","LP-13","LP-14","LP-15",
                  "LP-16","LP-17","LP-18","LP-19","LP-20","LP-21","LP-22","LP-23"}};

  private double[][] nodeData;
  private double[][] branchData;
  private double[][] genData;
  private double[][] lse3SecData;  // 3-sectional 8-hour LSE data
  private double[][] lseDataFixedDemand;  // Combine 3-sectional to 24-hour LSE data
  private double[][] lseDataPriceSensitiveDemand; 
  private boolean isPU;
  private static double baseS;
  private static double baseV;
  private static int demandType; // demandType = 0 if fixed demand; demandType = 1 if mixed demand

  private ArrayList slist = new ArrayList();
    //slist: the master string list to store tokenized strings in data file

  // Constructor
  public INIT(String initFilePath) {

    //Do an initial default load in case some parameters are missing in the configuration file.
    loadDefaultConfig();
    try {
      readInitData(initFilePath);
    }
    catch (Exception ioe) {loadDefaultConfig();
      System.err.println(
          "Problem loading simulation configurtation from file. " +
          "Using default configuration.");
    }
  }

  private void readInitData(String initFilePath){

    try{

      // Create a stream tokenizer by creating a buffer reader and a file reader
      FileReader       fr = new FileReader(initFilePath);
      BufferedReader   br = new BufferedReader(fr);
      StreamTokenizer tok = new StreamTokenizer(br);

      // Resets tokenizer's syntax table so that all characters are "ordinary"
      // to override Java's default setting for tokenization, which is too
      // finely tuned.
      tok.resetSyntax();

      // Set characters in the ASCII range 33 to 255 as valid word characters
      tok.wordChars(33,255);

      // Set characters in the ASCII range 0 to 32(' ') as whitespace characters
      tok.whitespaceChars(0,' ');   // or: tok.whitespaceChars(0,32);

    //Specified that the character argument starts a single-line comment.
      tok.commentChar('/');

      // Add tokens to a String list, slist
      while(tok.nextToken() != tok.TT_EOF){
        slist.add(tok.sval);
      }

      // Parse slist to double[][] arrays, The 'main' method
      parseStringList();

      //Close file reader
      fr.close();

    }catch(Exception e){
    System.out.println("Exception: " + e);
  }
  }

  private void parseStringList(){
    setDemandType();
    setNodeData();
    setBranchData();
    setGenData();
    setLSEDataFixedDemand();
    if(demandType == 1){
      setLSEDataPriceSensitiveDemand();
    }
      
    int index = slist.indexOf("UNIT")+1;
    isPU = slist.get(index).toString().equalsIgnoreCase("PU");
    if(isPU == false){
      convertToPerUnit();
    }

  }

  private double[][] correctRoundingError(double[][] a){
    for(int i=0; i<a.length; i++){
      for(int j=0; j<a[0].length; j++){
        for(int p=1; p<1.0E6; p=p*10){
          if(Math.abs(a[i][j]*p-Math.rint(a[i][j]*p)) < TOL){
            a[i][j] = Math.rint(a[i][j]*p)/p;
            break;
          }
        }
      }
    }
    return a;
  }

  private double correctRoundingError(double scalar){
    for(int p=1; p<1.0E6; p=p*10){
      if(Math.abs(scalar*p-Math.rint(scalar*p)) < TOL){
        scalar = Math.rint(scalar*p)/p;
        break;
      }
    }
    return scalar;
  }


  private void convertToPerUnit(){

    boolean isThereBaseS = slist.contains((String) "BASE_S");
    boolean isThereBaseV = slist.contains((String) "BASE_V");

    if(isThereBaseS == true){
      int index = slist.indexOf("BASE_S")+1;
      baseS = Double.parseDouble(slist.get(index).toString());
    }
    else{
      baseS = BASE_S;
    }
    if(isThereBaseV == true){
      int index = slist.indexOf("BASE_V")+1;
      baseV = Double.parseDouble(slist.get(index).toString());
    }
    else{
      baseV = BASE_V;
    }

    for(int n=0; n<branchData.length; n++){
      // Convert lineCap from SI to PU
      branchData[n][2] = branchData[n][2]/baseS;
      // Convert reactance from SI to PU, x(pu) = x/Zo = x/(Vo^2/So) = (x*So)/Vo^2
      branchData[n][3] = (branchData[n][3]*baseS)/(baseV*baseV);
    }
    for(int i=0; i<genData.length; i++){
      genData[i][3] = genData[i][3]*baseS;       // Convert a from SI to PU-adjusted
      genData[i][4] = genData[i][4]*baseS*baseS; // Convert b from SI to PU-adjusted
      genData[i][5] = genData[i][5]/baseS;       // Convert capL from SI to PU
      genData[i][6] = genData[i][6]/baseS;       // Convert capU from SI to PU
    }
    if(demandType == 1){
      for(int i=0; i<lseDataPriceSensitiveDemand.length; i++){
        lseDataPriceSensitiveDemand[i][2] = lseDataPriceSensitiveDemand[i][2]*baseS;       
              // Convert c from SI to PU-adjusted C
        lseDataPriceSensitiveDemand[i][3] = lseDataPriceSensitiveDemand[i][3]*baseS*baseS; 
              // Convert d from SI to PU-adjusted D
        lseDataPriceSensitiveDemand[i][4] = lseDataPriceSensitiveDemand[i][4]/baseS;       
              // Convert sLoadL from SI to PU-adjusted sLoadL
        lseDataPriceSensitiveDemand[i][5] = lseDataPriceSensitiveDemand[i][5]/baseS;       
              // Convert sLoadU from SI to PU-adjusted sLoadU
    }
    }
    for(int j=0; j<lseDataFixedDemand.length; j++){
      for(int h=2; h<lseDataFixedDemand[0].length; h++){
        lseDataFixedDemand[j][h] = lseDataFixedDemand[j][h]/baseS; // Convert LP from SI to PU
      }
    }
    branchData = correctRoundingError(branchData);
    genData    = correctRoundingError(genData);
    lseDataFixedDemand    = correctRoundingError(lseDataFixedDemand);
  }

  
  private void setDemandType(){
    int i = slist.indexOf("DemandType")+1;
    
    String demandTypeString = slist.get(i).toString();
    //System.out.println("Demand type in string: " + demandTypeString);
    if(demandTypeString.equalsIgnoreCase("PriceSensitiveNone")){
      demandType = 0;
    }
    else if(demandTypeString.equalsIgnoreCase("PriceSensitiveAll")){
      demandType = 1;
    }
    else{
      System.out.println("Wrong input for demand type!");
    }
    //demandType = Integer.parseInt(slist.get(i).toString());   
    //System.out.println("Demand type: " + demandType);
    //demandType = (int) Double.parseDouble(slist.get(i).toString());
  }
  
  private void setNodeData(){
    int fromIndex = slist.indexOf("NodeData")+1;
    int   toIndex = slist.indexOf("BranchData");
    int numValues = slist.subList(fromIndex,toIndex).size();
       //NOTE: fromIndex is inclusive, and toIndex is exclusive for subList().
    nodeData = new double[numValues/NUM_NODE_PARAMS][NUM_NODE_PARAMS];
    parseToArray2D(nodeData, NUM_NODE_PARAMS, fromIndex, toIndex);
  }

  private void setBranchData(){
    int fromIndex = slist.indexOf("BranchData")+1;
    int   toIndex = slist.indexOf("GenData");
    int numValues = slist.subList(fromIndex,toIndex).size();
    branchData = new double[numValues/NUM_BRANCH_PARAMS][NUM_BRANCH_PARAMS];
    parseToArray2D(branchData, NUM_BRANCH_PARAMS, fromIndex, toIndex);
  }

  private void setGenData(){
    int fromIndex = slist.indexOf("GenData")+1;
    int toIndex = -1;
    if (demandType == 0){
      toIndex = slist.indexOf("LSEDataFixedDemand");  
    }
    else if(demandType == 1){
      toIndex = slist.indexOf("LSEDataPriceSensitiveDemand");
    }
    else{
      System.out.println("Wrong demand type!");
    }
    int numValues = slist.subList(fromIndex,toIndex).size();
    genData = new double[numValues/NUM_GEN_PARAMS][NUM_GEN_PARAMS];
    parseToArray2D(genData, NUM_GEN_PARAMS, fromIndex, toIndex);
  }

  private void setLSEDataPriceSensitiveDemand(){
    int fromIndex = slist.indexOf("LSEDataPriceSensitiveDemand")+1;
    int   toIndex = slist.indexOf("LSEDataFixedDemand");
    int numValues = slist.subList(fromIndex,toIndex).size();
    lseDataPriceSensitiveDemand = 
        new double[numValues/NUM_LSE_PARAMS_SLOAD][NUM_LSE_PARAMS_SLOAD];
    parseToArray2D(lseDataPriceSensitiveDemand, NUM_LSE_PARAMS_SLOAD, 
        fromIndex, toIndex);
  }
  
  private void setLSEDataFixedDemand(){
    
    int fromIndex = slist.indexOf("LSEDataFixedDemand")+1;
    int   toIndex = slist.indexOf("EndOfFile");
    int numValues = slist.subList(fromIndex,toIndex).size();

    // Create lse3SecData as double[90/10][10] or 9-by-10 2D array
    int row3 = numValues/NUM_LSE_3SEC_PARAMS;    // row3 = 90/10 = 9
    int col3 = NUM_LSE_3SEC_PARAMS;              // col3 = 10
    lse3SecData = new double[row3][col3];
    parseToArray2D(lse3SecData, NUM_LSE_3SEC_PARAMS, fromIndex, toIndex);

    // Create lseData as double[90/(3*10)][26] or 3-by-26 2D array
    int row = numValues/(3*NUM_LSE_3SEC_PARAMS); // row = 90/(3*10) = 3
    int col = NUM_LSE_PARAMS;                    // col = 26
    lseDataFixedDemand = new double[row][col];

    DoubleMatrix2D lse3SecDataDM = new DenseDoubleMatrix2D(lse3SecData);
    DoubleMatrix2D lseDataDM = new DenseDoubleMatrix2D(lseDataFixedDemand);
    lseDataDM.viewPart(0,0,row,col3)
        .assign(lse3SecDataDM.viewPart(0,0,row,col3));
    lseDataDM.viewPart(0,col3,row,col3-2)
        .assign(lse3SecDataDM.viewPart(row,2,row,col3-2));
    lseDataDM.viewPart(0,2*col3-2,row,col3-2)
        .assign(lse3SecDataDM.viewPart(2*row,2,row,col3-2));
    lseDataFixedDemand = lseDataDM.copy().toArray();
  }

  private void parseToArray2D(double[][] data, int numParams,
                            int fromIndex, int toIndex){
    int row = 0;
    int col = -1;
    for(int i=fromIndex; i<toIndex; i++){
      if(col >= numParams-1){
        col = 0;
        row++;
      }
      else{
        col++;
      }
      data[row][col] = Double.parseDouble(slist.get(i).toString());
    }
  }

  private void loadDefaultConfig(){

  }

  public String[][] getParamNames(){
    return paramNames;
  }
  public static double getBaseS(){
    return baseS;
  }
  public static double getBaseV(){
     return baseV;
   }

  public static int getDemandType(){
    return demandType;
  }  
  public double[][] getNodeData(){
    return nodeData;
  }

  public double[][] getBranchData(){
    return branchData;
  }
  public double[][] getGenData(){
    return genData;
  }
  public double[][] getLSEDataPriceSensitiveDemand(){
    return lseDataPriceSensitiveDemand;
  }
  public double[][] getLSEDataFixedDemand(){
    return lseDataFixedDemand;
  }


  public static void printArray(double [][]a){
    for(int i = 0; i<a.length; i++){
      for(int j = 0 ; j<a[0].length; j++){
        System.out.print("\t" + a[i][j]);
        }
        System.out.println();
      }
      System.out.println();
  }

  public static void printArray(String[] s){
    for(int i=0; i<s.length; i++){
      System.out.print("\t" + s[i]);
    }
    System.out.println();
  }


}
